/*================================================================*/
/* COPYRIGHT:                                                     */
/* ----------                                                     */
/* Copyright (C) International Business Machines Corp., 2000.     */
/*                                                                */
/* DISCLAIMER OF WARRANTIES:                                      */
/* -------------------------                                      */
/* The following [enclosed] code is sample code created by IBM    */
/* Corporation.  This sample code is not part of any standard IBM */
/* product and is provided to you solely for the purpose of       */
/* assisting you in the development of your applications.  The    */
/* code is provided "AS IS", without warranty of any kind.  IBM   */
/* shall not be liable for any damages arising out of your use of */
/* the sample code, even if they have been advised of the         */
/* possibility of such damages.                                   */
/*================================================================*/

#include <windows.h>
#include <windowsx.h>
#include <mmsystem.h>
#include <mmreg.h>
#include <msacm.h>

#include <stdio.h>

#include <smrc.h>

#define FUNCTION(type) extern "C" __declspec( dllexport ) type __stdcall

// This is a Windows - only sample
// It uses

#define BYTE_ORDER      2		// This might be 1 or ?
#define BYTES_PER_BLOCK 4120            // This is whatever???

#define HEADER_SIZE     13*2		// This is unchangeable

// Set up a structure to hold all of the (cough) global data structures...

struct AudioInfo {
  int sample_rate;
  int loglevel;
  FILE *logfile;
  char *init_string;
  char *source;

  unsigned char *pbSrc;
  unsigned char *pbDst;
  unsigned long cbSrcLength, cbSrcLengthUsed;
  unsigned long cbDstLength, cbDstLengthUsed; 
  unsigned long nSrcBytesPerSec, nDstBytesPerSec;
  unsigned long cbSrcBlockSize;

  // windows crap
  HMMIO hmmio;
  HACMSTREAM has;
  
  AudioInfo(FILE *_logfile, int _loglevel) : 
    logfile(_logfile), loglevel(_loglevel), init_string(NULL), source(NULL),
    hmmio(NULL), has(NULL), pbSrc(NULL), pbDst(NULL)
  {}

  ~AudioInfo() {
    delete pbSrc;
    delete pbDst;
    if (hmmio) mmioClose(hmmio, 0);
    if (has) acmStreamClose(has, 0);
    delete init_string;
    delete source;
  }

};

static AudioInfo *ainfo = NULL;
static FILE *logfile;
static int loglevel;



// This gets a filename in value if we are doing file I/O

FUNCTION(int)
AudioSetSource( short length, void *value ) 
{
  // ignore length field, assume value is a string...
  delete ainfo->source;
  if (value) {
    ainfo->source = new char [strlen((const char *) value) + 1];
    strcpy(ainfo->source, (const char *) value);
  }
  else {
    ainfo->source = NULL;
  }
  return SM_RC_OK; 
}

FUNCTION(int)
AudioQuerySource(short *length, void **value) 
{
  if (ainfo->source) {
    *length = strlen(ainfo->source) + 1;
    value = (void **) &ainfo->source;
  } else {
    *length = 0;
    value = NULL;
  }
  return SM_RC_OK;
}



FUNCTION(int)
AudioConnect(const char *init_string, const char * /*client_name*/,
	int *handle, int sample_rate, int file_format, int *data_size,
        int *byte_order, const char ** /*signature*/, char * /*cepslan_data*/) 
{

  if (ainfo) { // We are already connected ?
    return SM_RC_BAD_AUDIO;
  }

  if (file_format != 0) {
    // we cannot handle non-PCM input or non-native byte order here...
    return SM_RC_BAD_AUDIO;
  }

  ainfo = new AudioInfo(logfile, loglevel);

  if (init_string) AudioSetSource(0,(void *) init_string);
  if (strcmp(init_string,"pcmfast")==0) { /*this is something special*/ }

  // init_string is not used can be "pcmfast" or whatever, make local copy
  if (init_string) {
    ainfo->init_string = new char [strlen(init_string)+1];
    strcpy(ainfo->init_string, init_string);
  }

  // client_name is a char * to client_name, ignore it...

  // handle is garbage, probably unitialized at this point, the
  // engine will fill in handle with a call go AudioGetHandle when it
  // returns from connect - set to garbage
  *handle = -1;

  // signature is never used by the engine, it is a unitialized pointer
  // capslan_data is not set if we are doing pcm

  ainfo->sample_rate = sample_rate;

  *data_size  = BYTES_PER_BLOCK-HEADER_SIZE;
  *byte_order = BYTE_ORDER;

  return SM_RC_OK;
}

// This stupid function returns a value which is never used ever!

FUNCTION(int)
AudioGetHandle( ) 
{
  return -1;
}

FUNCTION(int)
AudioDisconnect( )
{
  delete ainfo;
  ainfo = NULL;
  return SM_RC_OK;
}


FUNCTION(int)
AudioCreate(const char *filename, FILE *_logfile, int _loglevel) 
{
  logfile = _logfile;
  loglevel = _loglevel;
  return SM_RC_OK;
}


FUNCTION(void)
AudioDestroy( ) 
{
  AudioDisconnect();
  logfile = NULL;
  loglevel = 0;
}




// report errors is for reporting over-runs.  This can't happen with a file...

FUNCTION(int)
AudioStartRecording( unsigned long *start_time, int /*report_errors*/ )
{
  unsigned long dw;
  MMRESULT mmr;
  WAVEFORMATEX *pwfx = NULL, dstwfx;
  MMCKINFO ck, ckRIFF;
  *start_time = GetTickCount();

  // Open the file given by source for input and set up an ACM stream
  // conversion so that it is ready to pump PCM to the engine...

  dstwfx.wFormatTag = WAVE_FORMAT_PCM;
  dstwfx.nChannels = 1;
  dstwfx.nSamplesPerSec = ainfo->sample_rate;
  dstwfx.nAvgBytesPerSec = ainfo->sample_rate * 2;
  dstwfx.nBlockAlign = 2;
  dstwfx.wBitsPerSample = 16;
  dstwfx.cbSize = 0;

  if (ainfo->source == NULL) {
    fprintf(logfile, " - Audio - No file specified for input\n");
    fflush(logfile);
    return SM_RC_RECORD_OPEN_ERROR;
  }

  if (ainfo->hmmio != NULL) {
    // already have open file?  close it
    mmioClose(ainfo->hmmio, 0);
    ainfo->hmmio = NULL;
  }

  ainfo->hmmio = mmioOpen( ainfo->source, NULL, MMIO_READ | MMIO_ALLOCBUF);
  if (ainfo->hmmio == NULL) {
    fprintf(logfile, " - Audio - mmioOpen failed\n");
    goto error;
  }

  // locate the WAVE form type
  ckRIFF.fccType = mmioFOURCC('W','A','V','E');
  if ( mmioDescend(ainfo->hmmio, &ckRIFF, NULL, MMIO_FINDRIFF) ) {
    fprintf(logfile, " - Audio - mmioDescend failed for WAVE chunk\n");
    goto error;
  }

  // Take only the first WAVE chunk... 
  while (MMSYSERR_NOERROR == mmioDescend(ainfo->hmmio, &ck, &ckRIFF, 0)) {
    if ((ck.dwDataOffset + ck.cksize) > (ckRIFF.dwDataOffset + ckRIFF.cksize)) {
      // this is a corrupt RIFF file
      fprintf(logfile, " - Audio - corrupt WAVE chunk\n");
      goto error;
    }
    switch (ck.ckid) {
      case mmioFOURCC('f','m','t',' '):
        if (pwfx != NULL) break; // Another format chunk!?
        dw = ck.cksize;
        if (dw<sizeof(WAVEFORMATEX)) dw=sizeof(WAVEFORMATEX);
        pwfx = (WAVEFORMATEX * ) malloc(dw);
        if (mmioRead(ainfo->hmmio, (HPSTR) pwfx, ck.cksize) 
            != (long) ck.cksize) {
          fprintf(logfile, " - Audio - mmioRead failed for WAVEFORMAT chunk\n");
          goto error;
        }
        break;
      case mmioFOURCC('d','a','t','a'):
        // format must come before data in a WAVE file.  We will stay at this
        // point...  I apologize for the spaghetti logic here...
        goto datachunk;
 
      default: // skip chunks we do not understand...
        break;
    }
    mmioAscend(ainfo->hmmio, &ck, 0);
  }

  // We never found a data chunk or format!
  fprintf(logfile, " - Audio - Did not find a data chunk\n");
  goto error;

  // We are ready to stream data from mmioRead to the engine, must convert
  // to PCM Linear 16 using ACM...

datachunk:
  if (ainfo->has) acmStreamClose(ainfo->has, 0);
  ainfo->has = NULL;
  mmr = acmStreamOpen(
    &ainfo->has, NULL, pwfx, &dstwfx, NULL, 0L, 0L, 
    ACM_STREAMOPENF_NONREALTIME);
  if (mmr == ACMERR_NOTPOSSIBLE) {
    fprintf(logfile, " - Audio - Could not convert file to suitable format\n");
    goto error;
  }
  if (mmr != MMSYSERR_NOERROR) goto error; // Stream conversion

  // Allocate a 5 second buffer for Src and Dst conversion streams

  delete ainfo->pbSrc;
  ainfo->nSrcBytesPerSec = pwfx->nAvgBytesPerSec;
  ainfo->cbSrcLength = pwfx->nAvgBytesPerSec * 5;
  ainfo->cbSrcLengthUsed = 0;
  ainfo->cbSrcBlockSize = pwfx->nBlockAlign;
  ainfo->pbSrc = new unsigned char [ ainfo->cbSrcLength ];
  
  ainfo->nDstBytesPerSec = dstwfx.nAvgBytesPerSec;
  ainfo->cbDstLength = dstwfx.nAvgBytesPerSec * 5 ;
  ainfo->cbDstLengthUsed = 0;
  ainfo->pbDst = new unsigned char [ ainfo->cbDstLength ];
  
  free(pwfx);  // no longer need wave format of file...

  return SM_RC_OK;    

error:
  if (ainfo->has) acmStreamClose(ainfo->has, 0);
  ainfo->has = NULL;
  if (ainfo->hmmio) mmioClose(ainfo->hmmio,0);
  ainfo->hmmio = NULL;
  if (pwfx) free(pwfx);
  if (logfile) {
    fprintf(logfile, " - Audio - Could not open file %s", ainfo->source );
    fflush(logfile);
  }
  return SM_RC_RECORD_OPEN_ERROR;
}


FUNCTION(int)
AudioGetPCM( char *block_buffer, long max_bytes, 
		long *new_bytes, int *end_of_pcm )
{
  MMRESULT mmr;
  unsigned long i, j, p;

  *end_of_pcm = 1; // always assume we are ending unless otherwise wrong

  // read as much as possible from file...
  if (ainfo->cbSrcLengthUsed < ainfo->cbSrcLength) {
    mmr = mmioRead(ainfo->hmmio, (LPSTR) &ainfo->pbSrc[ainfo->cbSrcLengthUsed],
      ainfo->cbSrcLength - ainfo->cbSrcLengthUsed);
    if (mmr < 0) {
      fprintf(logfile, " - Audio - Error during mmioRead\n");
      goto error;
    }
    ainfo->cbSrcLengthUsed += mmr;
    if (mmr > 0) *end_of_pcm = 0; // more data in file
  }

  if (ainfo->cbSrcLengthUsed>=ainfo->cbSrcBlockSize) { // more source to convert
    ACMSTREAMHEADER ash;
    ash.cbStruct = sizeof(ACMSTREAMHEADER);
    ash.fdwStatus = 0L;
    ash.dwUser = 0L;
    ash.pbSrc = ainfo->pbSrc;
    ash.cbSrcLength = ainfo->cbSrcLengthUsed;
    ash.cbSrcLengthUsed = 0;
    ash.pbDst = &ainfo->pbDst[ainfo->cbDstLengthUsed];
    ash.cbDstLength = ainfo->cbDstLength-ainfo->cbDstLengthUsed;
    ash.cbDstLengthUsed = 0L;
    ash.dwDstUser = 0;
    ash.dwSrcUser = 0;

    // MS ACM's are sloppy, they to not watch the DstLength size,
    // so never send more source then there is room for
    unsigned long maxSrc = ash.cbDstLength * 
		ainfo->nSrcBytesPerSec / ainfo->nDstBytesPerSec;
    if (maxSrc < ash.cbSrcLength) ash.cbSrcLength = maxSrc;
    if (ash.cbSrcLength >= 0) {

      mmr = acmStreamPrepareHeader(ainfo->has, &ash, 0);
      if (mmr != MMSYSERR_NOERROR) {
        __asm int 3;
        fprintf(logfile, " - Audio - failure to prepare acmStreamHeader\n");
        goto error;
     }

      mmr = acmStreamConvert(ainfo->has, &ash, (*end_of_pcm == 0) ?
        ACM_STREAMCONVERTF_BLOCKALIGN : 0 );

      mmr |= acmStreamUnprepareHeader(ainfo->has, &ash, 0);
      if (mmr != MMSYSERR_NOERROR) {
        fprintf(logfile, " - Audio - failure to prepare acmStreamHeader\n");
        goto error;
      }
  
      // Okay, slam the unused part of the source over
      // ainfo->pbSrc[0] to ainfo->pbSrc[ash.cbSrcLengthUsed] was absorbed
      // ainfo->pbSrc[cbSrcLengthUsed] to ainfo->pbSrc[cbSrcLengthUsed]
      // is still needed
      for (j=0, i=ash.cbSrcLengthUsed; i<ainfo->cbSrcLengthUsed; i++) {
        ainfo->pbSrc[j++] = ainfo->pbSrc[i];
      }
      ainfo->cbSrcLengthUsed -= ash.cbSrcLengthUsed;
    }

    if (ash.cbSrcLengthUsed > 0) *end_of_pcm = 0; // more data was converted
    if (ash.cbDstLengthUsed > 0) *end_of_pcm = 0; // more output was generated
    
    ainfo->cbDstLengthUsed += ash.cbDstLengthUsed;
  }

  // Now pack as much of the dest buffer as we can into the buffers for 
  // the engine, adding the (puke) headers on the fly...

  for (p=0, i=0; i<(unsigned) max_bytes/BYTES_PER_BLOCK; i++) {
    // we do not have a complete block, only send if end_of_pcm
    if (*end_of_pcm==0 && 
        ainfo->cbDstLengthUsed - p < BYTES_PER_BLOCK-HEADER_SIZE) break;

    short *header = (short *) &block_buffer[i * BYTES_PER_BLOCK];
    header [ 0 ] = 0; // not used damn it.
    header [ 1 ] = 0; // not used ! (format?)
    header [ 2 ] = (BYTES_PER_BLOCK - HEADER_SIZE)/2; // samples
    header [ 3 ] = 0; // not used ! (volume?)
    header [ 4 ] = 0; // not used ! (Eddie's one)
    header [ 5 ] = 0; // not used ! (block index)
    header [ 6 ] = 0; // not used ! (end_of_pcm );
    header [ 7 ] = 0; // not used
    header [ 8 ] = 0; // not used
    header [ 9 ] = 0; // not used
    header [ 10 ] = 0; // not used
    header [ 11 ] = BYTES_PER_BLOCK - HEADER_SIZE;
    header [ 12 ] = BYTES_PER_BLOCK - HEADER_SIZE;

    for (j=HEADER_SIZE; j<BYTES_PER_BLOCK; j++) {
      if (p>=ainfo->cbDstLengthUsed) break;
      block_buffer[i * BYTES_PER_BLOCK + j] = ainfo->pbDst[p++];
    }

    // zero out extra data in the last block sent...
    // if we get here, we are on the last block ever...
    if (j<BYTES_PER_BLOCK) {
      for (;j<BYTES_PER_BLOCK;j++) block_buffer[i * BYTES_PER_BLOCK + j] =0;
      break;
    }
  }
  *new_bytes = i * BYTES_PER_BLOCK;

  // 
  // ainfo->pbDst[0] to ainfo->pbDst[p] was sent
  // ainfo->pbDst[p] to ainfo->pbDst[cbDstLengthUsed] keep for next time

  for (i=p; i<ainfo->cbDstLengthUsed; i++) {
    ainfo->pbDst[i-p] = ainfo->pbDst[i];
  }
  ainfo->cbDstLengthUsed -= p; 

  if (ainfo->cbDstLengthUsed > 0) *end_of_pcm = 0; // still have data to send
  return SM_RC_OK;

error:
  return SM_RC_BAD_AUDIO;
}

FUNCTION(int)
AudioStopRecording() 
{
  delete ainfo->pbSrc; ainfo->pbSrc = NULL;
  delete ainfo->pbDst; ainfo->pbDst = NULL;
  if (ainfo->has) acmStreamClose(ainfo->has, 0);
  ainfo->has = NULL;
  if (ainfo->hmmio) mmioClose(ainfo->hmmio,0);
  ainfo->hmmio = NULL;
  return SM_RC_OK;
}

FUNCTION(int)
AudioStartPlayback()
{ 
  return -1; 
}

FUNCTION(int)
AudioPutPCM( char *, long )
{
  return -1;
}

FUNCTION(int)
AudioStopPlayback( int )
{
  return -1;
}

// All of these funtions are totally vestigial and should be banished forever

FUNCTION(int) AudioQueryConfig( long *config ) { return (*config = 0); }
FUNCTION(int) AudioQueryDevices( long *devices ) { return (*devices = 0); }
FUNCTION(int) AudioSetDevice( long ) { return 0; }
FUNCTION(int) AudioSetInput( long ) { return 0; }
FUNCTION(int) AudioSetOutput( long ) { return 0; }
FUNCTION(int) AudioSetInputGain( long ) { return 0; }
FUNCTION(int) AudioSetOutputGain( long ) { return 0; }
